@@ -14,7 +14,7 @@
 --
 -- This module uses common names and so is designed to be imported qualified:
 --
--- > import qualified Codec.Archive.Tar.Index as Tar
+-- > import qualified Codec.Archive.Tar.Index as TarIndex
 --
 -----------------------------------------------------------------------------
 module Codec.Archive.Tar.Index (
@@ -28,6 +28,10 @@ module Codec.Archive.Tar.Index (
     -- @tar@ file is needed to 'build' the 'TarIndex', but thereafter you can
     -- 'lookup' paths in the @tar@ file, and then use 'hReadEntry' to
     -- seek to the right part of the file and read the entry.
+    --
+    -- An index cannot be used to lookup 'Directory' entries in a tar file;
+    -- instead, you will get 'TarDir' entry listing all the entries in the
+    -- directory.
 
     -- * Index type
     TarIndex,
@@ -35,6 +39,7 @@ module Codec.Archive.Tar.Index (
     -- * Index lookup
     lookup,
     TarIndexEntry(..),
+    toList,
 
     -- ** I\/O operations
     TarEntryOffset,
@@ -46,10 +51,11 @@ module Codec.Archive.Tar.Index (
     -- ** Incremental construction
     -- $incremental-construction
     IndexBuilder,
-    emptyIndex,
+    empty,
     addNextEntry,
     skipNextEntry,
-    finaliseIndex,
+    finalise,
+    unfinalise,
 
     -- * Serialising indexes
     serialise,
@@ -64,10 +70,18 @@ module Codec.Archive.Tar.Index (
     indexEndEntryOffset,
     indexNextEntryOffset,
 
+    -- * Deprecated aliases
+    emptyIndex,
+    finaliseIndex,
+
 #ifdef TESTS
     prop_lookup,
+    prop_toList,
     prop_valid,
-    prop_index_matches_tar
+    prop_serialise_deserialise,
+    prop_serialiseSize,
+    prop_index_matches_tar,
+    prop_finalise_unfinalise,
 #endif
   ) where
 
@@ -76,9 +90,9 @@ import Data.Typeable (Typeable)
 import Codec.Archive.Tar.Types as Tar
 import Codec.Archive.Tar.Read  as Tar
 import qualified Codec.Archive.Tar.Index.StringTable as StringTable
-import Codec.Archive.Tar.Index.StringTable (StringTable(..))
+import Codec.Archive.Tar.Index.StringTable (StringTable, StringTableBuilder)
 import qualified Codec.Archive.Tar.Index.IntTrie as IntTrie
-import Codec.Archive.Tar.Index.IntTrie (IntTrie(..))
+import Codec.Archive.Tar.Index.IntTrie (IntTrie, IntTrieBuilder)
 
 import qualified System.FilePath.Posix as FilePath
 import Data.Monoid (Monoid(..))
@@ -91,25 +105,32 @@ import Data.Bits
 import qualified Data.Array.Unboxed as A
 import Prelude hiding (lookup)
 import System.IO
-import Control.Exception (throwIO)
+import Control.Exception (assert, throwIO)
+import Control.DeepSeq
 
-import qualified Data.ByteString      as BS
-import qualified Data.ByteString.Lazy as LBS
-#if MIN_VERSION_bytestring(0,10,2)
-import Data.ByteString.Builder      as BS
+import qualified Data.ByteString        as BS
+import qualified Data.ByteString.Char8  as BS.Char8
+import qualified Data.ByteString.Lazy   as LBS
+import qualified Data.ByteString.Unsafe as BS
+#if MIN_VERSION_bytestring(0,10,2) || defined(MIN_VERSION_bytestring_builder)
+import Data.ByteString.Builder          as BS
+import Data.ByteString.Builder.Extra    as BS (toLazyByteStringWith,
+                                               untrimmedStrategy)
 #else
-import Data.ByteString.Lazy.Builder as BS
+import Data.ByteString.Lazy.Builder     as BS
+import Data.ByteString.Lazy.Builder.Extras as BS (toLazyByteStringWith,
+                                                  untrimmedStrategy)
 #endif
 
 #ifdef TESTS
 import qualified Prelude
-import Test.QuickCheck hiding (Result)
-import Test.QuickCheck.Property (Result, exception, succeeded)
+import Test.QuickCheck
+import Test.QuickCheck.Property (ioProperty)
 import Control.Applicative ((<$>), (<*>))
 import Control.Monad (unless)
-import Data.List (nub, sort, stripPrefix, isPrefixOf)
+import Data.List (nub, sort, sortBy, stripPrefix, isPrefixOf)
 import Data.Maybe
-import System.IO.Unsafe (unsafePerformIO)
+import Data.Function (on)
 import Control.Exception (SomeException, try)
 import Codec.Archive.Tar.Write          as Tar
 import qualified Data.ByteString.Handle as HBS
@@ -150,6 +171,9 @@ data TarIndex = TarIndex
 
   deriving (Eq, Show, Typeable)
 
+instance NFData TarIndex where
+  rnf (TarIndex _ _ _) = () -- fully strict by construction
+
 -- | The result of 'lookup' in a 'TarIndex'. It can either be a file directly,
 -- or a directory entry containing further entries (and all subdirectories
 -- recursively). Note that the subtrees are constructed lazily, so it's
@@ -197,8 +221,9 @@ lookup (TarIndex pathTable pathTrie _) path = do
 toComponentIds :: StringTable PathComponentId -> FilePath -> Maybe [PathComponentId]
 toComponentIds table =
     lookupComponents []
-  . filter (/= ".")
-  . FilePath.splitDirectories
+  . filter (/= BS.Char8.singleton '.')
+  . splitDirectories
+  . BS.Char8.pack
   where
     lookupComponents cs' []     = Just (reverse cs')
     lookupComponents cs' (c:cs) = case StringTable.lookup table c of
@@ -206,17 +231,28 @@ toComponentIds table =
       Just cid -> lookupComponents (cid:cs') cs
 
 fromComponentId :: StringTable PathComponentId -> PathComponentId -> FilePath
-fromComponentId table = StringTable.index table
+fromComponentId table = BS.Char8.unpack . StringTable.index table
+
+-- | All the files in the index with their corresponding 'TarEntryOffset's.
+--
+-- Note that the files are in no special order. If you intend to read all or
+-- most files then is is recommended to sort by the 'TarEntryOffset'.
+--
+toList :: TarIndex -> [(FilePath, TarEntryOffset)]
+toList (TarIndex pathTable pathTrie _) =
+    [ (path, off)
+    | (cids, off) <- IntTrie.toList pathTrie
+    , let path = FilePath.joinPath (map (fromComponentId pathTable) cids) ]
 
 
 -- | Build a 'TarIndex' from a sequence of tar 'Entries'. The 'Entries' are
 -- assumed to start at offset @0@ within a file.
 --
 build :: Entries e -> Either e TarIndex
-build = go emptyIndex
+build = go empty
   where
     go !builder (Next e es) = go (addNextEntry e builder) es
-    go !builder  Done       = Right $! finaliseIndex builder
+    go !builder  Done       = Right $! finalise builder
     go !_       (Fail err)  = Left err
 
 
@@ -224,59 +260,71 @@ build = go emptyIndex
 -- If you need more control than 'build' then you can construct the index
 -- in an acumulator style using the 'IndexBuilder' and operations.
 --
--- Start with the 'emptyIndex' and use 'addNextEntry' (or 'skipNextEntry') for
+-- Start with 'empty' and use 'addNextEntry' (or 'skipNextEntry') for
 -- each 'Entry' in the tar file in order. Every entry must added or skipped in
 -- order, otherwise the resulting 'TarIndex' will report the wrong
--- 'TarEntryOffset's. At the end use 'finaliseIndex' to get the 'TarIndex'.
+-- 'TarEntryOffset's. At the end use 'finalise' to get the 'TarIndex'.
 --
 -- For example, 'build' is simply:
 --
--- > build = go emptyIndex
+-- > build = go empty
 -- >   where
 -- >     go !builder (Next e es) = go (addNextEntry e builder) es
--- >     go !builder  Done       = Right $! finaliseIndex builder
+-- >     go !builder  Done       = Right $! finalise builder
 -- >     go !_       (Fail err)  = Left err
 
 
 -- | The intermediate type used for incremental construction of a 'TarIndex'.
 --
-data IndexBuilder = IndexBuilder [(FilePath, TarEntryOffset)]
-                                 {-# UNPACK #-} !TarEntryOffset
+data IndexBuilder
+   = IndexBuilder !(StringTableBuilder PathComponentId)
+                  !(IntTrieBuilder PathComponentId TarEntryOffset)
+   {-# UNPACK #-} !TarEntryOffset
+  deriving (Eq, Show)
+
+instance NFData IndexBuilder where
+  rnf (IndexBuilder _ _ _) = () -- fully strict by construction
 
 -- | The initial empty 'IndexBuilder'.
 --
+empty :: IndexBuilder
+empty = IndexBuilder StringTable.empty IntTrie.empty 0
+
 emptyIndex :: IndexBuilder
-emptyIndex = IndexBuilder [] 0
+emptyIndex = empty
+{-# DEPRECATED emptyIndex "Use TarIndex.empty" #-}
 
 -- | Add the next 'Entry' into the 'IndexBuilder'.
 --
 addNextEntry :: Entry -> IndexBuilder -> IndexBuilder
-addNextEntry entry (IndexBuilder acc nextOffset) =
-    IndexBuilder ((entrypath, nextOffset):acc)
+addNextEntry entry (IndexBuilder stbl itrie nextOffset) =
+    IndexBuilder stbl' itrie'
                  (nextEntryOffset entry nextOffset)
   where
-    !entrypath  = entryPath entry
+    !entrypath    = splitTarPath (entryTarPath entry)
+    (stbl', cids) = StringTable.inserts entrypath stbl
+    itrie'        = IntTrie.insert cids nextOffset itrie
 
 -- | Use this function if you want to skip some entries and not add them to the
 -- final 'TarIndex'.
 --
 skipNextEntry :: Entry -> IndexBuilder -> IndexBuilder
-skipNextEntry entry (IndexBuilder acc nextOffset) =
-    IndexBuilder acc (nextEntryOffset entry nextOffset)
+skipNextEntry entry (IndexBuilder stbl itrie nextOffset) =
+    IndexBuilder stbl itrie (nextEntryOffset entry nextOffset)
 
 -- | Finish accumulating 'Entry' information and build the compact 'TarIndex'
 -- lookup structure.
 --
-finaliseIndex :: IndexBuilder -> TarIndex
-finaliseIndex (IndexBuilder pathsOffsets finalOffset) =
+finalise :: IndexBuilder -> TarIndex
+finalise (IndexBuilder stbl itrie finalOffset) =
     TarIndex pathTable pathTrie finalOffset
   where
-    pathComponents = concatMap (FilePath.splitDirectories . fst) pathsOffsets
-    pathTable = StringTable.construct pathComponents
-    pathTrie  = IntTrie.construct
-                  [ (cids, offset)
-                  | (path, offset) <- pathsOffsets
-                  , let Just cids = toComponentIds pathTable path ]
+    pathTable = StringTable.finalise stbl
+    pathTrie  = IntTrie.finalise itrie
+
+finaliseIndex :: IndexBuilder -> TarIndex
+finaliseIndex = finalise
+{-# DEPRECATED finaliseIndex "Use TarIndex.finalise" #-}
 
 -- | This is the offset immediately following the entry most recently added
 -- to the 'IndexBuilder'. You might use this if you need to know the offsets
@@ -284,7 +332,7 @@ finaliseIndex (IndexBuilder pathsOffsets finalOffset) =
 -- Use with 'hSeekEntryOffset'. See also 'nextEntryOffset'.
 --
 indexNextEntryOffset :: IndexBuilder -> TarEntryOffset
-indexNextEntryOffset (IndexBuilder _ off) = off
+indexNextEntryOffset (IndexBuilder _ _ off) = off
 
 -- | This is the offset immediately following the last entry in the tar file.
 -- This can be useful to append further entries into the tar file.
@@ -308,10 +356,43 @@ nextEntryOffset entry offset =
       OtherEntryType _ _ size -> blocks size
       _                       -> 0
   where
-    -- NOTE: The special case for 0 is important to avoid underflow,
-    -- because we are computing an unsigned TarEntryOffset (aka Word32) value
-    blocks 0    = 0
-    blocks size = 1 + ((fromIntegral size - 1) `div` 512)
+    -- NOTE: to avoid underflow, do the (fromIntegral :: Int64 -> Word32) last
+    blocks :: Int64 -> TarEntryOffset
+    blocks size = fromIntegral (1 + (size - 1) `div` 512)
+
+type FilePathBS = BS.ByteString
+
+splitTarPath :: TarPath -> [FilePathBS]
+splitTarPath (TarPath name prefix) =
+    splitDirectories prefix ++ splitDirectories name
+
+splitDirectories :: FilePathBS -> [FilePathBS]
+splitDirectories bs =
+    case BS.Char8.split '/' bs of
+      c:cs | BS.null c -> BS.Char8.singleton '/' : filter (not . BS.null) cs
+      cs               ->                          filter (not . BS.null) cs
+
+
+-------------------------
+-- Resume building an existing index
+--
+
+-- | Resume building an existing index
+--
+-- A 'TarIndex' is optimized for a highly compact and efficient in-memory
+-- representation. This, however, makes it read-only. If you have an existing
+-- 'TarIndex' for a large file, and want to add to it, you can translate the
+-- 'TarIndex' back to an 'IndexBuilder'. Be aware that this is a relatively
+-- costly operation (linear in the size of the 'TarIndex'), though still
+-- faster than starting again from scratch.
+--
+-- This is the left inverse to 'finalise' (modulo ordering).
+--
+unfinalise :: TarIndex -> IndexBuilder
+unfinalise (TarIndex pathTable pathTrie finalOffset) =
+    IndexBuilder (StringTable.unfinalise pathTable)
+                 (IntTrie.unfinalise pathTrie)
+                 finalOffset
 
 
 -------------------------
@@ -460,85 +541,59 @@ hSeekEndEntryOffset hnd Nothing = do
 -- | The 'TarIndex' is compact in memory, and it has a similarly compact
 -- external representation.
 --
-serialise :: TarIndex -> BS.Builder
-serialise (TarIndex stringTable intTrie finalOffset) =
-     BS.word32BE 1 -- format version
+serialise :: TarIndex -> BS.ByteString
+serialise = toStrict . serialiseLBS
+
+-- we keep this version around just so we can check we got the size right.
+serialiseLBS :: TarIndex -> LBS.ByteString
+serialiseLBS index =
+    BS.toLazyByteStringWith
+      (BS.untrimmedStrategy (serialiseSize index) 512) LBS.empty
+      (serialiseBuilder index)
+
+serialiseSize :: TarIndex -> Int
+serialiseSize (TarIndex stringTable intTrie _) =
+    StringTable.serialiseSize stringTable
+  + IntTrie.serialiseSize intTrie
+  + 8
+
+serialiseBuilder :: TarIndex -> BS.Builder
+serialiseBuilder (TarIndex stringTable intTrie finalOffset) =
+     BS.word32BE 2 -- format version
   <> BS.word32BE finalOffset
-  <> serialiseStringTable stringTable
-  <> serialiseIntTrie intTrie
+  <> StringTable.serialise stringTable
+  <> IntTrie.serialise intTrie
 
 -- | Read the external representation back into a 'TarIndex'.
 --
 deserialise :: BS.ByteString -> Maybe (TarIndex, BS.ByteString)
 deserialise bs
-  | BS.length bs >= 8
-  , let ver = readWord32BE bs 0
+  | BS.length bs < 8
+  = Nothing
+
+  | let ver = readWord32BE bs 0
   , ver == 1
   = do let !finalOffset = readWord32BE bs 4
-       (stringTable, bs')  <- deserialiseStringTable (BS.drop 8 bs)
-       (intTrie,     bs'') <- deserialiseIntTrie bs'
+       (stringTable, bs')  <- StringTable.deserialiseV1 (BS.drop 8 bs)
+       (intTrie,     bs'') <- IntTrie.deserialise bs'
+       return (TarIndex stringTable intTrie finalOffset, bs'')
+
+  | let ver = readWord32BE bs 0
+  , ver == 2
+  = do let !finalOffset = readWord32BE bs 4
+       (stringTable, bs')  <- StringTable.deserialiseV2 (BS.drop 8 bs)
+       (intTrie,     bs'') <- IntTrie.deserialise bs'
        return (TarIndex stringTable intTrie finalOffset, bs'')
 
   | otherwise = Nothing
 
-serialiseIntTrie :: IntTrie k v -> BS.Builder
-serialiseIntTrie (IntTrie arr) =
-    let (_, !ixEnd) = A.bounds arr in
-    BS.word32BE (ixEnd+1)
- <> foldr (\n r -> BS.word32BE n <> r) mempty (A.elems arr)
-
-deserialiseIntTrie :: BS.ByteString -> Maybe (IntTrie k v, BS.ByteString)
-deserialiseIntTrie bs
-  | BS.length bs >= 4
-  , let lenArr   = readWord32BE bs 0
-        lenTotal = 4 + 4 * fromIntegral lenArr
-  , BS.length bs >= 4 + 4 * fromIntegral lenArr
-  , let !arr = A.array (0, lenArr-1)
-                      [ (i, readWord32BE bs off)
-                      | (i, off) <- zip [0..lenArr-1] [4,8 .. lenTotal - 4] ]
-        !bs' = BS.drop lenTotal bs
-  = Just (IntTrie arr, bs')
-
-  | otherwise
-  = Nothing
-
-serialiseStringTable :: StringTable id -> BS.Builder
-serialiseStringTable (StringTable strs arr) =
-      let (_, !ixEnd) = A.bounds arr in
-
-      BS.word32BE (fromIntegral (BS.length strs))
-   <> BS.word32BE (fromIntegral ixEnd + 1)
-   <> BS.byteString strs
-   <> foldr (\n r -> BS.word32BE n <> r) mempty (A.elems arr)
-
-deserialiseStringTable :: BS.ByteString -> Maybe (StringTable id, BS.ByteString)
-deserialiseStringTable bs
-  | BS.length bs >= 8
-  , let lenStrs = fromIntegral (readWord32BE bs 0)
-        lenArr  = fromIntegral (readWord32BE bs 4)
-        lenTotal= 8 + lenStrs + 4 * lenArr
-  , BS.length bs >= lenTotal
-  , let strs = BS.take lenStrs (BS.drop 8 bs)
-        arr  = A.array (0, lenArr-1)
-                       [ (i, readWord32BE bs off)
-                       | (i, off) <- zip [0..lenArr-1]
-                                         [offArrS,offArrS+4 .. offArrE]
-                       ]
-        offArrS = 8 + lenStrs
-        offArrE = offArrS + 4 * lenArr - 1
-        !stringTable = StringTable strs arr
-        !bs'         = BS.drop lenTotal bs
-  = Just (stringTable, bs')
-
-  | otherwise
-  = Nothing
-
 readWord32BE :: BS.ByteString -> Int -> Word32
 readWord32BE bs i =
-     fromIntegral (BS.index bs (i + 0)) `shiftL` 24
-   + fromIntegral (BS.index bs (i + 1)) `shiftL` 16
-   + fromIntegral (BS.index bs (i + 2)) `shiftL` 8
-   + fromIntegral (BS.index bs (i + 3))
+    assert (i >= 0 && i+3 <= BS.length bs - 1) $
+    fromIntegral (BS.unsafeIndex bs (i + 0)) `shiftL` 24
+  + fromIntegral (BS.unsafeIndex bs (i + 1)) `shiftL` 16
+  + fromIntegral (BS.unsafeIndex bs (i + 2)) `shiftL` 8
+  + fromIntegral (BS.unsafeIndex bs (i + 3))
 
 
 -------------------------
@@ -553,17 +608,24 @@ readWord32BE bs i =
 prop_lookup :: ValidPaths -> NonEmptyFilePath -> Bool
 prop_lookup (ValidPaths paths) (NonEmptyFilePath p) =
   case (lookup index p, Prelude.lookup p paths) of
-    (Nothing,                    Nothing)      -> True
-    (Just (TarFileEntry offset), Just offset') -> offset == offset'
-    (Just (TarDir entries),      Nothing)      -> sort (nub (map fst entries))
-                                               == sort (nub completions)
-    _                                          -> False
+    (Nothing,                    Nothing)          -> True
+    (Just (TarFileEntry offset), Just (_,offset')) -> offset == offset'
+    (Just (TarDir entries),      Nothing)          -> sort (nub (map fst entries))
+                                                   == sort (nub completions)
+    _                                              -> False
   where
-    index       = finaliseIndex (IndexBuilder paths 0)
+    index       = construct paths
     completions = [ head (FilePath.splitDirectories completion)
                   | (path,_) <- paths
                   , completion <- maybeToList $ stripPrefix (p ++ "/") path ]
 
+prop_toList :: ValidPaths -> Bool
+prop_toList (ValidPaths paths) =
+    sort (toList index)
+ == sort [ (path, off) | (path, (_sz, off)) <- paths ]
+  where
+    index = construct paths
+
 prop_valid :: ValidPaths -> Bool
 prop_valid (ValidPaths paths)
   | not $ StringTable.prop_valid   pathbits = error "TarIndex: bad string table"
@@ -573,39 +635,67 @@ prop_valid (ValidPaths paths)
   | otherwise                               = True
 
   where
-    index@(TarIndex pathTable _ _) = finaliseIndex (IndexBuilder paths 0)
+    index@(TarIndex pathTable _ _) = construct paths
 
-    pathbits = concatMap (FilePath.splitDirectories . fst) paths
+    pathbits = concatMap (map BS.Char8.pack . FilePath.splitDirectories . fst)
+                         paths
     intpaths = [ (cids, offset)
-               | (path, offset) <- paths
+               | (path, (_size, offset)) <- paths
                , let Just cids = toComponentIds pathTable path ]
-    prop' = flip all paths $ \(file, offset) ->
+    prop' = flip all paths $ \(file, (_size, offset)) ->
       case lookup index file of
         Just (TarFileEntry offset') -> offset' == offset
         _                           -> False
 
+prop_serialise_deserialise :: ValidPaths -> Bool
+prop_serialise_deserialise (ValidPaths paths) =
+    Just (index, BS.empty) == (deserialise . serialise) index
+  where
+    index = construct paths
+
+prop_serialiseSize :: ValidPaths -> Bool
+prop_serialiseSize (ValidPaths paths) =
+    case (LBS.toChunks . serialiseLBS) index of
+      [c1] -> BS.length c1 == serialiseSize index
+      _    -> False
+  where
+    index = construct paths
+
 newtype NonEmptyFilePath = NonEmptyFilePath FilePath deriving Show
 
 instance Arbitrary NonEmptyFilePath where
   arbitrary = NonEmptyFilePath . FilePath.joinPath
                 <$> listOf1 (elements ["a", "b", "c", "d"])
 
-newtype ValidPaths = ValidPaths [(FilePath, TarEntryOffset)] deriving Show
+newtype ValidPaths = ValidPaths [(FilePath, (Int64, TarEntryOffset))] deriving Show
 
 instance Arbitrary ValidPaths where
-  arbitrary =
-      ValidPaths . makeNoPrefix <$> listOf ((,) <$> arbitraryPath <*> arbitrary)
+  arbitrary = do
+      paths <- makeNoPrefix <$> listOf arbitraryPath
+      sizes <- vectorOf (length paths) (getNonNegative <$> arbitrary)
+      let offsets = scanl (\o sz -> o + 1 + blocks sz) 0 sizes
+      return (ValidPaths (zip paths (zip sizes offsets)))
     where
       arbitraryPath   = FilePath.joinPath
                          <$> listOf1 (elements ["a", "b", "c", "d"])
       makeNoPrefix [] = []
-      makeNoPrefix ((k,v):kvs)
-        | all (\(k', _) -> not (isPrefixOfOther k k')) kvs
-                     = (k,v) : makeNoPrefix kvs
-        | otherwise  =         makeNoPrefix kvs
+      makeNoPrefix (k:ks)
+        | all (not . isPrefixOfOther k) ks
+                     = k : makeNoPrefix ks
+        | otherwise  =     makeNoPrefix ks
 
       isPrefixOfOther a b = a `isPrefixOf` b || b `isPrefixOf` a
 
+      blocks :: Int64 -> TarEntryOffset
+      blocks size = fromIntegral (1 + ((size - 1) `div` 512))
+
+-- Helper for bulk construction.
+construct :: [(FilePath, (Int64, TarEntryOffset))] -> TarIndex
+construct =
+    either (\_ -> undefined) id
+  . build
+  . foldr (\(path, (size, _off)) es -> Next (testEntry path size) es) Done
+
 example0 :: Entries ()
 example0 =
          testEntry "foo-1.0/foo-1.0.cabal" 1500 -- at block 0
@@ -632,11 +722,10 @@ data SimpleTarArchive = SimpleTarArchive {
 instance Show SimpleTarArchive where
   show = show . simpleTarRaw
 
-prop_index_matches_tar :: SimpleTarArchive -> Result
+prop_index_matches_tar :: SimpleTarArchive -> Property
 prop_index_matches_tar sta =
-    case unsafePerformIO (try go) of
-      Left ex  -> exception "" ex
-      Right () -> succeeded
+    ioProperty (try go >>= either (\e -> throwIO (e :: SomeException))
+                                  (\_ -> return True))
   where
     go :: IO ()
     go = do
@@ -701,6 +790,33 @@ instance Arbitrary SimpleTarArchive where
         ,  510 ,  511 ,  512 ,  513 ,  514
         , 1022 , 1023 , 1024 , 1025 , 1026
         ]
+
+-- | 'IndexBuilder' constructed from a 'SimpleIndex'
+newtype SimpleIndexBuilder = SimpleIndexBuilder IndexBuilder
+  deriving Show
+
+instance Arbitrary SimpleIndexBuilder where
+  arbitrary = SimpleIndexBuilder . build' . simpleTarEntries <$> arbitrary
+    where
+      -- like 'build', but don't finalize
+      build' :: Show e => Entries e -> IndexBuilder
+      build' = go empty
+        where
+          go !builder (Next e es) = go (addNextEntry e builder) es
+          go !builder  Done       = builder
+          go !_       (Fail err)  = error (show err)
+
+prop_finalise_unfinalise :: SimpleIndexBuilder -> Bool
+prop_finalise_unfinalise (SimpleIndexBuilder index) =
+    unfinalise (finalise index) == index
+
+#endif
+
+toStrict :: LBS.ByteString -> BS.ByteString
+#if MIN_VERSION_bytestring(0,10,0)
+toStrict = LBS.toStrict
+#else
+toStrict = BS.concat . LBS.toChunks
 #endif
 
 #if !(MIN_VERSION_base(4,5,0))
